@@ -68,15 +68,20 @@ static ulonglong timespec_to_usec(const struct timespec *ts)
   return (ulonglong) ts->tv_sec * TIME_MILLION + ts->tv_nsec / TIME_THOUSAND;
 }
 
-int signal_waiting_transaction(THD *waiting_thd, const char *binlog_file,
-                                my_off_t binlog_pos)
+static int
+signal_waiting_transaction(THD *waiting_thd, bool thd_valid,
+                           const char *binlog_file, my_off_t binlog_pos)
 {
   /*
     It is possible that the connection thd waiting for an ACK was killed. In
     such circumstance, the connection thread will nullify the thd member of its
     Active_tranx node. So before we try to signal, ensure the THD exists.
+
+    The thd_valid is only set while the THD is waiting in commit_trx(); this
+    is defensive coding to not signal an invalid THD if we somewhere
+    accidentally did not remove the transaction from the list.
   */
-  if (waiting_thd)
+  if (waiting_thd && thd_valid)
     mysql_cond_signal(&waiting_thd->COND_wakeup_ready);
   return 0;
 }
@@ -182,6 +187,7 @@ int Active_tranx::insert_tranx_node(THD *thd_to_wait,
   ins_node->log_name[FN_REFLEN-1] = 0; /* make sure it ends properly */
   ins_node->log_pos = log_file_pos;
   ins_node->thd= thd_to_wait;
+  ins_node->thd_valid= false;
 
   if (!m_trx_front)
   {
@@ -263,7 +269,8 @@ void Active_tranx::clear_active_tranx_nodes(
     if ((log_file_name != NULL) &&
         compare(new_front, log_file_name, log_file_pos) > 0)
       break;
-    pre_delete_hook(new_front->thd, new_front->log_name, new_front->log_pos);
+    pre_delete_hook(new_front->thd, new_front->thd_valid,
+                    new_front->log_name, new_front->log_pos);
     new_front = new_front->next;
   }
 
@@ -355,13 +362,17 @@ void Active_tranx::unlink_thd_as_waiter(const char *log_file_name,
   }
 
   if (entry)
+  {
     entry->thd= NULL;
+    entry->thd_valid= false;
+  }
 
   DBUG_VOID_RETURN;
 }
 
-bool Active_tranx::is_thd_waiter(THD *thd_to_check, const char *log_file_name,
-                                 my_off_t log_file_pos)
+Tranx_node *
+Active_tranx::is_thd_waiter(THD *thd_to_check, const char *log_file_name,
+                            my_off_t log_file_pos)
 {
   DBUG_ENTER("Active_tranx::assert_thd_is_waiter");
   mysql_mutex_assert_owner(m_lock);
@@ -377,7 +388,7 @@ bool Active_tranx::is_thd_waiter(THD *thd_to_check, const char *log_file_name,
     entry = entry->hash_next;
   }
 
-  DBUG_RETURN(static_cast<bool>(entry));
+  DBUG_RETURN(entry);
 }
 
 /*******************************************************************************
@@ -565,12 +576,14 @@ void Repl_semi_sync_master::remove_slave()
 {
   lock();
   DBUG_ASSERT(rpl_semi_sync_master_clients > 0);
-  if (!(--rpl_semi_sync_master_clients) && !rpl_semi_sync_master_wait_no_slave)
+  if (!(--rpl_semi_sync_master_clients) && !rpl_semi_sync_master_wait_no_slave
+      && get_master_enabled())
   {
     /*
       Signal transactions waiting in commit_trx() that they do not have to
       wait anymore.
     */
+    DBUG_ASSERT(m_active_tranxs);
     m_active_tranxs->clear_active_tranx_nodes(NULL, 0,
                                               signal_waiting_transaction);
   }
@@ -608,13 +621,17 @@ int Repl_semi_sync_master::report_reply_packet(uint32 server_id,
       DBUG_RETURN(-1);
     }
     else
-      sql_print_error("Read semi-sync reply magic number error");
+      sql_print_error("Read semi-sync reply magic number error. "
+                      "Got magic: %u  command %u  length: %lu",
+                      (uint) packet[REPLY_MAGIC_NUM_OFFSET], (uint) packet[0],
+                      packet_len);
     goto l_end;
   }
 
   if (unlikely(packet_len < REPLY_BINLOG_NAME_OFFSET))
   {
-    sql_print_error("Read semi-sync reply length error: packet is too small");
+    sql_print_error("Read semi-sync reply length error: packet is too small: %lu",
+                    packet_len);
     goto l_end;
   }
 
@@ -622,7 +639,8 @@ int Repl_semi_sync_master::report_reply_packet(uint32 server_id,
   log_file_len = packet_len - REPLY_BINLOG_NAME_OFFSET;
   if (unlikely(log_file_len >= FN_REFLEN))
   {
-    sql_print_error("Read semi-sync reply binlog file length too large");
+    sql_print_error("Read semi-sync reply binlog file length too large: %llu",
+                    (ulonglong) log_file_pos);
     goto l_end;
   }
   strncpy(log_file_name, (const char*)packet + REPLY_BINLOG_NAME_OFFSET, log_file_len);
@@ -856,6 +874,10 @@ int Repl_semi_sync_master::commit_trx(const char *trx_wait_binlog_name,
 
   if (!rpl_semi_sync_master_clients && !rpl_semi_sync_master_wait_no_slave)
   {
+    lock();
+    m_active_tranxs->unlink_thd_as_waiter(trx_wait_binlog_name,
+                                          trx_wait_binlog_pos);
+    unlock();
     rpl_semi_sync_master_no_transactions++;
     DBUG_RETURN(0);
   }
@@ -915,6 +937,9 @@ int Repl_semi_sync_master::commit_trx(const char *trx_wait_binlog_name,
         }
       }
 
+      Tranx_node *tranx_entry=
+        m_active_tranxs->is_thd_waiter(thd, trx_wait_binlog_name,
+                                       trx_wait_binlog_pos);
       /* In between the binlogging of this transaction and this wait, it is
        * possible that our entry in Active_tranx was removed (i.e. if
        * semi-sync was switched off and on). It is also possible that the
@@ -925,16 +950,15 @@ int Repl_semi_sync_master::commit_trx(const char *trx_wait_binlog_name,
        * rpl_semi_sync_master_yes/no_tx consistent with it, we check for a
        * semi-sync restart _after_ checking the reply state.
        */
-      if (unlikely(!m_active_tranxs->is_thd_waiter(thd, trx_wait_binlog_name,
-                                                   trx_wait_binlog_pos)))
+      if (unlikely(!tranx_entry))
       {
         DBUG_EXECUTE_IF(
             "semisync_log_skip_trx_wait",
             sql_print_information(
                 "Skipping semi-sync wait for transaction at pos %s, %lu. This "
                 "should be because semi-sync turned off and on during the "
-                "lifetime of this transaction.",
-                trx_wait_binlog_name, trx_wait_binlog_pos););
+                "lifetime of this transaction.", trx_wait_binlog_name,
+                static_cast<unsigned long>(trx_wait_binlog_pos)););
 
         /* The only known reason for a missing entry at this point is if
          * semi-sync was turned off then on, so on debug builds, we track
@@ -945,6 +969,16 @@ int Repl_semi_sync_master::commit_trx(const char *trx_wait_binlog_name,
         break;
       }
 
+      /*
+        Mark that our THD is now valid for signalling to by the ack thread.
+        It is important to ensure that we can never leave a no longer valid
+        THD in the transaction list and signal it, eg. MDEV-36934. This way,
+        we ensure the THD will only be signalled while this function is
+        running, even in case of some incorrect error handling or similar
+        that might leave a dangling THD in the list.
+      */
+      tranx_entry->thd_valid= true;
+
       /* Let us update the info about the minimum binlog position of waiting
        * threads.
        */
@@ -1277,6 +1311,8 @@ int Repl_semi_sync_master::write_tranx_in_binlog(THD *thd,
 
   DBUG_ENTER("Repl_semi_sync_master::write_tranx_in_binlog");
 
+  DEBUG_SYNC(current_thd, "semisync_at_write_tranx_in_binlog");
+
   lock();
 
   /* This is the real check inside the mutex. */
@@ -1310,7 +1346,8 @@ int Repl_semi_sync_master::write_tranx_in_binlog(THD *thd,
     m_commit_file_name_inited = true;
   }
 
-  if (is_on())
+  if (is_on() &&
+      (rpl_semi_sync_master_clients || rpl_semi_sync_master_wait_no_slave))
   {
     DBUG_ASSERT(m_active_tranxs != NULL);
     if(m_active_tranxs->insert_tranx_node(thd, log_file_name, log_file_pos))
@@ -1446,6 +1483,22 @@ void Repl_semi_sync_master::set_export_stats()
   unlock();
 }
 
+void Repl_semi_sync_master::reset_stats()
+{
+  lock();
+  rpl_semi_sync_master_yes_transactions = 0;
+  rpl_semi_sync_master_no_transactions = 0;
+  rpl_semi_sync_master_off_times = 0;
+  rpl_semi_sync_master_timefunc_fails = 0;
+  rpl_semi_sync_master_wait_sessions = 0;
+  rpl_semi_sync_master_wait_pos_backtraverse = 0;
+  rpl_semi_sync_master_trx_wait_num = 0;
+  rpl_semi_sync_master_trx_wait_time = 0;
+  rpl_semi_sync_master_net_wait_num = 0;
+  rpl_semi_sync_master_net_wait_time = 0;
+  unlock();
+}
+
 void Repl_semi_sync_master::await_all_slave_replies(const char *msg)
 {
   struct timespec timeout;
@@ -1468,7 +1521,7 @@ void Repl_semi_sync_master::await_all_slave_replies(const char *msg)
     if (msg && first)
     {
       first= false;
-      sql_print_information(msg);
+      sql_print_information("%s", msg);
     }
 
     wait_result=
